/*
** Module   : tty.keyboard.js
** Abstract : keystrokes management.
**
** Copyright (c) 2014, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**         Golden Code Development Corporation
**                 CONFIDENTIAL
**
** -#- -I- --Date-- ----------------------Description--------------------------
** 001 MAG 20140110 First version.
*/

tty.keyboard = (function(socket, tty)
{
   /* exports object */
   var my = {}; 

   /* key constants */   
   var keys = 
   {
      BACKSPACE: 8,
      TAB: 9,
      CLEAR: 12,
      ENTER: 13,
      SHIFT: 16,
      CTRL: 17,
      ALT: 18,
      PAUSE: 19,
      CAPS_LOCK: 20,
      ESCAPE: 27,
      SPACE: 32,
      PAGE_UP: 33,
      PAGE_DOWN: 34,
      END: 35,
      HOME: 36,
      LEFT_ARROW: 37,
      UP_ARROW: 38,
      RIGHT_ARROW: 39,
      DOWN_ARROW: 40,
      INSERT: 45,
      DELETE: 46,
      HELP: 47,
      LEFT_WINDOW: 91,
      RIGHT_WINDOW: 92,
      SELECT: 93,
      NUMPAD_0: 96,
      NUMPAD_1: 97,
      NUMPAD_2: 98,
      NUMPAD_3: 99,
      NUMPAD_4: 100,
      NUMPAD_5: 101,
      NUMPAD_6: 102,
      NUMPAD_7: 103,
      NUMPAD_8: 104,
      NUMPAD_9: 105,
      NUMPAD_MULTIPLY: 106,
      NUMPAD_PLUS: 107,
      NUMPAD_ENTER: 108,
      NUMPAD_MINUS: 109,
      NUMPAD_PERIOD: 110,
      NUMPAD_DIVIDE: 111,
      F1: 112,
      F2: 113,
      F3: 114,
      F4: 115,
      F5: 116,
      F6: 117,
      F7: 118,
      F8: 119,
      F9: 120,
      F10: 121,
      F11: 122,
      F12: 123,
      F13: 124,
      F14: 125,
      F15: 126,
      NUM_LOCK: 144,
      SCROLL_LOCK: 145,
      UP_DPAD: 175,
      DOWN_DPAD: 176,
      LEFT_DPAD: 177,
      RIGHT_DPAD: 178
   };            

   /* normal key mapping */
   var xlate = {};

   /* ALT key mapping */
   var xlateAlt = {}; 
   
   /* track escape status */
   var escape = false;
             
   /* initialize module */
   my.init = function()
   {      
      /* keypress */
      document.addEventListener("keypress", function(evt)
      {
         var key = evt.charCode;
         if (key != 0)
         {
            sendKey(key, evt);
         }         
      });

      /* keydown */
      document.addEventListener("keydown", function(evt)
      {
         var key = evt.keyCode;

         if (key == keys.ESCAPE)
         {
            escape = !escape;
            return;
         }
         
         if (key == keys.CTRL || key == keys.ALT || key == keys.SHIFT)
         {
            return;
         }
                      
         var p4code = -1;
         var charCode = mapUSkeys(evt);
                                          
         // Handle ALT or Escape
         if (evt.altKey || escape)
         {
            p4code = xlateAlt[key];
            
            if (p4code === undefined)
            {
               if (charCode != -1 && ((!escape && evt.altKey) || (escape && !evt.altKey)))
               {
                  p4code = processCtrl(evt, charCode);
                  if(p4code == -1)
                  {
                     p4code = charCode;
                  }
                  p4code |= 1024;
               }
               else
               {
                  p4code = -1;
               }
            }
            else
            {
               if (!escape && evt.altKey && p4code == keys.DELETE)
               {
                  // special case for ALT+DELETE key combo
                  p4code = 512;
               }            
            }    
                       
            escape = false;
         }
         // Normal
         else
         {
            /* normal map */
            var p4code = xlate[key];
            
            if (p4code === undefined)
            {
               // check CTRL
               p4code = processCtrl(evt, charCode);
            }
            else
            {
               if (p4code == keys.TAB && evt.shiftKey)
               {
                  p4code = 509;
               }                           
            }
         }                                                           

         if (p4code != -1)
         {
            sendKey(p4code, evt);
         }
      });
      
      // Normal keys
      xlate[keys.BACKSPACE] = 8;
      xlate[keys.TAB] = 9;
      xlate[keys.ENTER] = 13;
      xlate[keys.UP_ARROW] = 501;
      xlate[keys.DOWN_ARROW] = 502;
      xlate[keys.RIGHT_ARROW] = 503;
      xlate[keys.LEFT_ARROW] = 504;
      xlate[keys.HOME] = 505;
      xlate[keys.END] = 506;
      xlate[keys.PAGE_UP] = 507;
      xlate[keys.PAGE_DOWN] = 508;
      xlate[keys.INSERT] = 510;
      xlate[keys.DELETE] = 127;     
      xlate[keys.F1] = 301;
      xlate[keys.F2] = 302;
      xlate[keys.F3] = 303;
      xlate[keys.F4] = 304;
      xlate[keys.F5] = 305;
      xlate[keys.F6] = 306;
      xlate[keys.F7] = 307;
      xlate[keys.F8] = 308;
      xlate[keys.F9] = 309;
      xlate[keys.F10] = 310;
      xlate[keys.F11] = 311;
      xlate[keys.F12] = 312;
      
      // ALT-keys
      xlateAlt[keys.UP_ARROW] = 559;      
      xlateAlt[keys.DOWN_ARROW] = 518;
      xlateAlt[keys.RIGHT_ARROW] = 553;
      xlateAlt[keys.LEFT_ARROW] = 538;
      xlateAlt[keys.PAGE_UP] = 548;
      xlateAlt[keys.PAGE_DOWN] = 543;
      xlateAlt[keys.INSERT] = 533;
      xlateAlt[keys.DELETE] = 127;     
      xlateAlt[keys.HOME] = 528;
      xlateAlt[keys.END] = 523;
      xlateAlt[keys.F1] = 313;
      xlateAlt[keys.F2] = 314;
      xlateAlt[keys.F3] = 315;
      xlateAlt[keys.F4] = 316;
      xlateAlt[keys.F5] = 317;
      xlateAlt[keys.F6] = 318;
      xlateAlt[keys.F7] = 319;
      xlateAlt[keys.F8] = 320;
      xlateAlt[keys.F9] = 321;
      xlateAlt[keys.F10] = 322;
      xlateAlt[keys.F11] = 323;
      xlateAlt[keys.F12] = 324;
                                          
   };//init
                     
   /* process CTRL */
   var processCtrl = function(evt, code)
   {
      var ctrlCode = -1;
      
      if (evt.ctrlKey)
      {         
         if (code == 50) //'2'
         {
            ctrlCode = 0;
         }
         else if (code == 54)  //'6'
         {
            ctrlCode = 30;
         }
         else if (code == 45) //'-'
         {
            ctrlCode = 31;
         }
         else if (code == 92) //'\'
         {           
            ctrlCode = 28;       
         }
         else if (code == 93) //']'
         {           
            ctrlCode = 29;       
         }
         else
         {
            if (code >= 32 && code < 127)
            {
               ctrlCode = code & 0xffbf;
               // VK_ESCAPE - NOT DELIVERED, 4GL doesn't ever return this.                        
               if (ctrlCode == keys.ESCAPE)
               {
                  ctrlCode = -1;
               }
            }
         }
      }
      
      return ctrlCode;
   };
               
   /* send keyEvent */
   var sendKey = function(key, evt)
   {
      var data = {};
      data.c = 0;
      data.k = key;
      socket.send(JSON.stringify(data));
      evt.preventDefault();
      evt.stopPropagation();
   };

   /* map printable keys US layout */   
   var mapUSkeys = function(evt)
   {
      var key = evt.keyCode;
      
      // A...Z
      if (key >= 65 && key <= 90)
      {
         return key;
      }
      
      if (tty.isFirefox)
      {
         switch(key)
         {
            case 59:
               key = 186; //; : 
               break;
            case 107:
               key = 187; //= + 
               break;
            case 109:
               key = 189; //- _
               break;
         }
      }
            
      if (evt.shiftKey)
      {
         switch(key)
         {
            case 48:
               return 41; //)
            case 49:
               return 33; //!
            case 50:
               return 64; //@
            case 51:
               return 35; //#
            case 52:
               return 36; //$
            case 53:
               return 37; //%
            case 54:
               return 94; //^
            case 55:
               return 38; //&
            case 56:
               return 42; //*
            case 57:
               return 40; //(
            case 188:
               return 60; //<
            case 190:
               return 62; //>
            case 191:
               return 63; //?
            case 192:
               return 126; //~
            case 219:
               return 123; //{
            case 220:
               return 124; //|
            case 221:
               return 125; //}
            case 222:
               return 34; //"
            case 186:
               return 58; //:
            case 187:
               return 43; //+
            case 189:
               return 95; //_
         }
      }
      else
      {
         // 0...9
         if (key >= 48 && key <= 57)
         {
            return key;
         }

         switch(key)
         {
            case 188:
               return 44; //,
            case 190:
               return 46; //.
            case 191:
               return 47; ///
            case 192:
               return 96; //`
            case 219:
               return 91; //[
            case 220:
               return 92; //\
            case 221:
               return 93; //]
            case 222:
               return 34; //'
            case 186:
               return 59; //;
            case 187:
               return 61; //=
            case 189:
               return 45; //-
            // numpad
            case 96:
            case 97:
            case 98:
            case 99:
            case 100:
            case 101:
            case 102:
            case 103:
            case 104:
            case 105:
               return key - 48; //0...9
            case 106:
               return 42; //*
            case 107:
               return 43; //+
            case 109:
               return 45; //-
            case 110:
               return 46; //.
            case 111:
               return 47; ///
         }
      }
            
      return -1;
   };
                  
   /* exports */
   return my;   
    
}(tty.socket, tty));
